@@ -4,6 +4,8 @@
 //! A library for working with [Google Breakpad][breakpad]'s
 //! text-format [symbol files][symbolfiles].
 //!
+//! See the [walker][] module for documentation on CFI evaluation.
+//!
 //! The highest-level API provided by this crate is to use the
 //! [`Symbolizer`][symbolizer] struct.
 //!
@@ -15,41 +17,56 @@
 //!
 //! ```
 //! # std::env::set_current_dir(env!("CARGO_MANIFEST_DIR"));
-//! use breakpad_symbols::{SimpleSymbolSupplier,Symbolizer,SimpleFrame,SimpleModule};
+//! use breakpad_symbols::{SimpleSymbolSupplier, Symbolizer, SimpleFrame, SimpleModule};
+//! use debugid::DebugId;
 //! use std::path::PathBuf;
-//! let paths = vec!(PathBuf::from("../testdata/symbols/"));
-//! let supplier = SimpleSymbolSupplier::new(paths);
-//! let symbolizer = Symbolizer::new(supplier);
+//! use std::str::FromStr;
 //!
-//! // Simple function name lookup with debug file, debug id, address.
-//! assert_eq!(symbolizer.get_symbol_at_address("test_app.pdb",
-//!                                             "5A9832E5287241C1838ED98914E9B7FF1",
-//!                                             0x1010)
-//!               .unwrap(),
-//!               "vswprintf");
+//! #[tokio::main]
+//! async fn main() {
+//!     let paths = vec!(PathBuf::from("../testdata/symbols/"));
+//!     let supplier = SimpleSymbolSupplier::new(paths);
+//!     let symbolizer = Symbolizer::new(supplier);
+//!
+//!     // Simple function name lookup with debug file, debug id, address.
+//!     let debug_id = DebugId::from_str("5A9832E5287241C1838ED98914E9B7FF1").unwrap();
+//!     assert_eq!(symbolizer.get_symbol_at_address("test_app.pdb", debug_id, 0x1010)
+//!         .await
+//!         .unwrap(),
+//!         "vswprintf");
+//! }
 //! ```
 
-use failure::Error;
+use async_trait::async_trait;
+use debugid::{CodeId, DebugId};
 use log::{debug, trace, warn};
-use reqwest::blocking::Client;
-use reqwest::Url;
+use reqwest::{Client, Url};
+use tempfile::NamedTempFile;
 
 use std::borrow::Cow;
 use std::boxed::Box;
-use std::cell::RefCell;
 use std::collections::HashMap;
-use std::fmt;
 use std::fs;
-use std::io::{self, Read, Write};
+use std::io::{self, Write};
 use std::path::{Path, PathBuf};
+use std::sync::Mutex;
 use std::time::Duration;
 
 pub use minidump_common::traits::Module;
+pub use sym_file::walker;
 
 pub use crate::sym_file::{CfiRules, SymbolFile};
 
 mod sym_file;
 
+// Re-exports for the purposes of the cfi_eval fuzzer. Not public API.
+#[doc(hidden)]
+#[cfg(feature = "fuzz")]
+pub mod fuzzing_private_exports {
+    pub use crate::sym_file::walker::{eval_win_expr_for_fuzzer, walk_with_stack_cfi};
+    pub use crate::sym_file::{StackInfoWin, WinStackThing};
+}
+
 /// Statistics on the symbols of a module.
 #[derive(Default, Debug)]
 pub struct SymbolStats {
@@ -73,9 +90,9 @@ pub struct SimpleModule {
     pub base_address: Option<u64>,
     pub size: Option<u64>,
     pub code_file: Option<String>,
-    pub code_identifier: Option<String>,
+    pub code_identifier: Option<CodeId>,
     pub debug_file: Option<String>,
-    pub debug_id: Option<String>,
+    pub debug_id: Option<DebugId>,
     pub version: Option<String>,
 }
 
@@ -83,10 +100,10 @@ impl SimpleModule {
     /// Create a `SimpleModule` with the given `debug_file` and `debug_id`.
     ///
     /// Uses `default` for the remaining fields.
-    pub fn new(debug_file: &str, debug_id: &str) -> SimpleModule {
+    pub fn new(debug_file: &str, debug_id: DebugId) -> SimpleModule {
         SimpleModule {
             debug_file: Some(String::from(debug_file)),
-            debug_id: Some(String::from(debug_id)),
+            debug_id: Some(debug_id),
             ..SimpleModule::default()
         }
     }
@@ -104,16 +121,14 @@ impl Module for SimpleModule {
             .as_ref()
             .map_or(Cow::from(""), |s| Cow::Borrowed(&s[..]))
     }
-    fn code_identifier(&self) -> Cow<str> {
-        self.code_identifier
-            .as_ref()
-            .map_or(Cow::from(""), |s| Cow::Borrowed(&s[..]))
+    fn code_identifier(&self) -> Option<CodeId> {
+        self.code_identifier.as_ref().cloned()
     }
     fn debug_file(&self) -> Option<Cow<str>> {
         self.debug_file.as_ref().map(|s| Cow::Borrowed(&s[..]))
     }
-    fn debug_identifier(&self) -> Option<Cow<str>> {
-        self.debug_id.as_ref().map(|s| Cow::Borrowed(&s[..]))
+    fn debug_identifier(&self) -> Option<DebugId> {
+        self.debug_id
     }
     fn version(&self) -> Option<Cow<str>> {
         self.version.as_ref().map(|s| Cow::Borrowed(&s[..]))
@@ -158,14 +173,14 @@ fn replace_or_add_extension(filename: &str, match_extension: &str, new_extension
 ///
 /// [module_line]: https://chromium.googlesource.com/breakpad/breakpad/+/master/docs/symbol_files.md#MODULE-records
 /// [packagesymbols]: https://gist.github.com/luser/2ad32d290f224782fcfc#file-packagesymbols-py
-pub fn relative_symbol_path(module: &dyn Module, extension: &str) -> Option<String> {
+pub fn relative_symbol_path(module: &(dyn Module + Sync), extension: &str) -> Option<String> {
     module.debug_file().and_then(|debug_file| {
         module.debug_identifier().map(|debug_id| {
             // Can't use PathBuf::file_name here, it doesn't handle
             // Windows file paths on non-Windows.
             let leaf = leafname(&debug_file);
             let filename = replace_or_add_extension(leaf, "pdb", extension);
-            [leaf, &debug_id[..], &filename[..]].join("/")
+            [leaf, &debug_id.breakpad().to_string(), &filename[..]].join("/")
         })
     })
 }
@@ -190,14 +205,20 @@ pub fn relative_symbol_path(module: &dyn Module, extension: &str) -> Option<Stri
 /// Since symbol files can be on the order of a gigabyte(!) and downloaded
 /// from the network, aggressive caching is pretty important. The current
 /// approach is a nice balance of simple and effective.
-#[derive(Debug)]
+#[derive(Debug, thiserror::Error)]
 pub enum SymbolError {
     /// Symbol file could not be found.
     ///
     /// In this case other symbol providers may still be able to find it!
+    #[error("symbol file not found")]
     NotFound,
+    /// The module was lacking either the debug file or debug id, as such the
+    /// path of the symbol could not be generated.
+    #[error("the debug file or id were missing")]
+    MissingDebugFileOrId,
     /// Symbol file could not be loaded into memory.
-    LoadError(Error),
+    #[error("couldn't read input stream")]
+    LoadError(#[from] std::io::Error),
     /// Symbol file was too corrupt to be parsed at all.
     ///
     /// Because symbol files are pretty modular, many corruptions/ambiguities
@@ -205,7 +226,8 @@ pub enum SymbolError {
     /// (e.g. a bad STACK WIN line can be discarded without affecting anything
     /// else). But sometimes we can't make any sense of the symbol file, and
     /// you find yourself here.
-    ParseError(Error),
+    #[error("parse error: {0} at line {1}")]
+    ParseError(&'static str, u64),
 }
 
 /// An error produced by fill_symbol.
@@ -231,28 +253,20 @@ impl PartialEq for SymbolError {
             (self, other),
             (SymbolError::NotFound, SymbolError::NotFound)
                 | (SymbolError::LoadError(_), SymbolError::LoadError(_))
-                | (SymbolError::ParseError(_), SymbolError::ParseError(_))
+                | (SymbolError::ParseError(..), SymbolError::ParseError(..))
         )
     }
 }
 
-impl fmt::Display for SymbolError {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self {
-            SymbolError::NotFound => write!(f, "Not found"),
-            SymbolError::LoadError(e) => write!(f, "Load error: {}", e),
-            SymbolError::ParseError(e) => write!(f, "Parse error: {}", e),
-        }
-    }
-}
-
 /// A trait for things that can locate symbols for a given module.
+#[async_trait]
 pub trait SymbolSupplier {
     /// Locate and load a symbol file for `module`.
     ///
     /// Implementations may use any strategy for locating and loading
     /// symbols.
-    fn locate_symbols(&self, module: &dyn Module) -> Result<SymbolFile, SymbolError>;
+    async fn locate_symbols(&self, module: &(dyn Module + Sync))
+        -> Result<SymbolFile, SymbolError>;
 }
 
 /// An implementation of `SymbolSupplier` that loads Breakpad text-format symbols from local disk
@@ -273,8 +287,12 @@ impl SimpleSymbolSupplier {
     }
 }
 
+#[async_trait]
 impl SymbolSupplier for SimpleSymbolSupplier {
-    fn locate_symbols(&self, module: &dyn Module) -> Result<SymbolFile, SymbolError> {
+    async fn locate_symbols(
+        &self,
+        module: &(dyn Module + Sync),
+    ) -> Result<SymbolFile, SymbolError> {
         if let Some(rel_path) = relative_symbol_path(module, "sym") {
             for path in self.paths.iter() {
                 let test_path = path.join(&rel_path);
@@ -302,8 +320,12 @@ impl StringSymbolSupplier {
     }
 }
 
+#[async_trait]
 impl SymbolSupplier for StringSymbolSupplier {
-    fn locate_symbols(&self, module: &dyn Module) -> Result<SymbolFile, SymbolError> {
+    async fn locate_symbols(
+        &self,
+        module: &(dyn Module + Sync),
+    ) -> Result<SymbolFile, SymbolError> {
         if let Some(symbols) = self.modules.get(&*module.code_file()) {
             return SymbolFile::from_bytes(symbols.as_bytes());
         }
@@ -373,8 +395,7 @@ impl HttpSymbolSupplier {
     }
 }
 
-/// Save the data in `contents` to `path`.
-fn save_contents(contents: &[u8], url: &Url, final_path: &Path, tmp_path: &Path) -> io::Result<()> {
+fn create_cache_file(tmp_path: &Path, final_path: &Path) -> io::Result<NamedTempFile> {
     // Use tempfile to save things to our cache to ensure proper
     // atomicity of writes. We may want multiple instances of rust-minidump
     // to be sharing a cache, and we don't want one instance to see another
@@ -393,9 +414,10 @@ fn save_contents(contents: &[u8], url: &Url, final_path: &Path, tmp_path: &Path)
     })?;
     fs::create_dir_all(&base)?;
 
-    let mut temp = tempfile::NamedTempFile::new_in(tmp_path)?;
-    temp.write_all(contents)?;
+    NamedTempFile::new_in(tmp_path)
+}
 
+fn commit_cache_file(mut temp: NamedTempFile, final_path: &Path, url: &Url) -> io::Result<()> {
     // Append any extra metadata we also want to be cached as "INFO" lines,
     // because this is an established format that parsers will ignore the
     // contents of by default.
@@ -408,52 +430,96 @@ fn save_contents(contents: &[u8], url: &Url, final_path: &Path, tmp_path: &Path)
     // If another process already wrote this entry, prefer their value to
     // avoid needless file system churn.
     temp.persist_noclobber(final_path)?;
+
     Ok(())
 }
 
 /// Fetch a symbol file from the URL made by combining `base_url` and `rel_path` using `client`,
 /// save the file contents under `cache` + `rel_path` and also return them.
-fn fetch_symbol_file(
+async fn fetch_symbol_file(
     client: &Client,
     base_url: &Url,
-    rel_path: &str,
+    module: &(dyn Module + Sync),
     cache: &Path,
     tmp: &Path,
-) -> Result<(Url, Vec<u8>), Error> {
-    let url = base_url.join(rel_path)?;
-    debug!("Trying {}", url);
-    let mut res = client.get(url.clone()).send()?.error_for_status()?;
-    let mut buf = vec![];
-    res.read_to_end(&mut buf)?;
-    let local = cache.join(rel_path);
+) -> Result<SymbolFile, SymbolError> {
+    // This function is a bit of a complicated mess because we want to write
+    // the input to our symbol cache, but we're a streaming parser. So we
+    // use the bare SymbolFile::parse to get access to the contents of
+    // the input stream as it's downloaded+parsed to write it to disk.
+    //
+    // Note that caching is strictly "optional" because it's more important
+    // to parse the symbols. So if at any point the caching i/o fails, we just
+    // give up on caching but let the parse+download continue.
 
-    match save_contents(&buf, &url, &local, tmp) {
-        Ok(_) => {}
-        Err(e) => warn!("Failed to save symbol file in local disk cache: {}", e),
+    // First try to GET the file from a server
+    let rel_path = relative_symbol_path(module, "sym").ok_or(SymbolError::MissingDebugFileOrId)?;
+    let mut url = base_url
+        .join(&rel_path)
+        .map_err(|_| SymbolError::NotFound)?;
+    let code_id = module.code_identifier().unwrap_or_default();
+    url.query_pairs_mut()
+        .append_pair("code_id", code_id.as_str())
+        .append_pair("code_file", &module.code_file());
+    debug!("Trying {}", url);
+    let res = client
+        .get(url.clone())
+        .send()
+        .await
+        .and_then(|res| res.error_for_status())
+        .map_err(|_| SymbolError::NotFound)?;
+
+    // Now try to create the temp cache file (not yet in the cache)
+    let final_cache_path = cache.join(rel_path);
+    let mut temp = create_cache_file(tmp, &final_cache_path)
+        .map_err(|e| {
+            warn!("Failed to save symbol file in local disk cache: {}", e);
+        })
+        .ok();
+
+    // Now stream parse the file as it downloads.
+    let mut symbol_file = SymbolFile::parse_async(res, |data| {
+        // While we're downloading+parsing, save this data to the the disk cache too
+        if let Some(file) = temp.as_mut() {
+            if let Err(e) = file.write_all(data) {
+                // Give up on caching this.
+                warn!("Failed to save symbol file in local disk cache: {}", e);
+                temp = None;
+            }
+        }
+    })
+    .await?;
+    // Make note of what URL this symbol file was downloaded from.
+    symbol_file.url = Some(url.to_string());
+
+    // Try to finish the cache file and atomically swap it into the cache.
+    if let Some(temp) = temp {
+        let _ = commit_cache_file(temp, &final_cache_path, &url).map_err(|e| {
+            warn!("Failed to save symbol file in local disk cache: {}", e);
+        });
     }
-    Ok((url, buf))
+
+    Ok(symbol_file)
 }
 
+#[async_trait]
 impl SymbolSupplier for HttpSymbolSupplier {
-    fn locate_symbols(&self, module: &dyn Module) -> Result<SymbolFile, SymbolError> {
+    async fn locate_symbols(
+        &self,
+        module: &(dyn Module + Sync),
+    ) -> Result<SymbolFile, SymbolError> {
         // Check local paths first.
-        let local_result = self.local.locate_symbols(module);
+        let local_result = self.local.locate_symbols(module).await;
         if !matches!(local_result, Err(SymbolError::NotFound)) {
             // Everything but NotFound prevents cascading
             return local_result;
         }
         // Now try urls
-        if let Some(rel_path) = relative_symbol_path(module, "sym") {
-            for url in &self.urls {
-                if let Ok((full_url, buf)) =
-                    fetch_symbol_file(&self.client, url, &rel_path, &self.cache, &self.tmp)
-                {
-                    // Parse the symbol file and set the "url" statistic
-                    return SymbolFile::from_bytes(&buf).map(|mut file| {
-                        file.url = Some(full_url.into());
-                        file
-                    });
-                }
+        for url in &self.urls {
+            if let Ok(file) =
+                fetch_symbol_file(&self.client, url, module, &self.cache, &self.tmp).await
+            {
+                return Ok(file);
             }
         }
         // If we get this far, we have failed to find anything
@@ -475,6 +541,8 @@ pub trait FrameSymbolizer {
 pub trait FrameWalker {
     /// Get the instruction address that we're trying to unwind from.
     fn get_instruction(&self) -> u64;
+    /// Check whether the callee has a callee of its own.
+    fn has_grand_callee(&self) -> bool;
     /// Get the number of bytes the callee's callee's parameters take up
     /// on the stack (or 0 if unknown/invalid). This is needed for
     /// STACK WIN unwinding.
@@ -541,13 +609,13 @@ impl FrameSymbolizer for SimpleFrame {
 
 // Can't make Module derive Hash, since then it can't be used as a trait
 // object (because the hash method is generic), so this is a hacky workaround.
-type ModuleKey = (String, String, Option<String>, Option<String>);
+type ModuleKey = (String, Option<String>, Option<String>, Option<String>);
 
 /// Helper for deriving a hash key from a `Module` for `Symbolizer`.
-fn key(module: &dyn Module) -> ModuleKey {
+fn key(module: &(dyn Module + Sync)) -> ModuleKey {
     (
         module.code_file().to_string(),
-        module.code_identifier().to_string(),
+        module.code_identifier().map(|s| s.to_string()),
         module.debug_file().map(|s| s.to_string()),
         module.debug_identifier().map(|s| s.to_string()),
     )
@@ -574,21 +642,21 @@ fn key(module: &dyn Module) -> ModuleKey {
 /// [fill_symbol]: struct.Symbolizer.html#method.fill_symbol
 pub struct Symbolizer {
     /// Symbol supplier for locating symbols.
-    supplier: Box<dyn SymbolSupplier + 'static>,
+    supplier: Box<dyn SymbolSupplier + Send + Sync + 'static>,
     /// Cache of symbol locating results.
     // TODO?: use lru-cache: https://crates.io/crates/lru-cache/
     // note that using an lru-cache would mess up the fact that we currently
     // use this for statistics collection. Splitting out statistics would be
     // way messier but not impossible.
-    symbols: RefCell<HashMap<ModuleKey, Result<SymbolFile, SymbolError>>>,
+    symbols: Mutex<HashMap<ModuleKey, Result<SymbolFile, SymbolError>>>,
 }
 
 impl Symbolizer {
     /// Create a `Symbolizer` that uses `supplier` to locate symbols.
-    pub fn new<T: SymbolSupplier + 'static>(supplier: T) -> Symbolizer {
+    pub fn new<T: SymbolSupplier + Send + Sync + 'static>(supplier: T) -> Symbolizer {
         Symbolizer {
             supplier: Box::new(supplier),
-            symbols: RefCell::new(HashMap::new()),
+            symbols: Mutex::new(HashMap::new()),
         }
     }
 
@@ -601,15 +669,15 @@ impl Symbolizer {
     /// See [the module-level documentation][module] for an example.
     ///
     /// [module]: index.html
-    pub fn get_symbol_at_address(
+    pub async fn get_symbol_at_address(
         &self,
         debug_file: &str,
-        debug_id: &str,
+        debug_id: DebugId,
         address: u64,
     ) -> Option<String> {
         let k = (debug_file, debug_id);
         let mut frame = SimpleFrame::with_instruction(address);
-        self.fill_symbol(&k, &mut frame).ok()?;
+        self.fill_symbol(&k, &mut frame).await.ok()?;
         frame.function
     }
 
@@ -625,31 +693,39 @@ impl Symbolizer {
     ///
     /// ```
     /// # std::env::set_current_dir(env!("CARGO_MANIFEST_DIR"));
+    /// use std::str::FromStr;
+    /// use debugid::DebugId;
     /// use breakpad_symbols::{SimpleSymbolSupplier,Symbolizer,SimpleFrame,SimpleModule};
-    /// use std::path::PathBuf;
-    /// let paths = vec!(PathBuf::from("../testdata/symbols/"));
-    /// let supplier = SimpleSymbolSupplier::new(paths);
-    /// let symbolizer = Symbolizer::new(supplier);
-    /// let m = SimpleModule::new("test_app.pdb", "5A9832E5287241C1838ED98914E9B7FF1");
-    /// let mut f = SimpleFrame::with_instruction(0x1010);
-    /// let _ = symbolizer.fill_symbol(&m, &mut f);
-    /// assert_eq!(f.function.unwrap(), "vswprintf");
-    /// assert_eq!(f.source_file.unwrap(), r"c:\program files\microsoft visual studio 8\vc\include\swprintf.inl");
-    /// assert_eq!(f.source_line.unwrap(), 51);
+    ///
+    /// #[tokio::main]
+    /// async fn main() {
+    ///     use std::path::PathBuf;
+    ///     let paths = vec!(PathBuf::from("../testdata/symbols/"));
+    ///     let supplier = SimpleSymbolSupplier::new(paths);
+    ///     let symbolizer = Symbolizer::new(supplier);
+    ///     let debug_id = DebugId::from_str("5A9832E5287241C1838ED98914E9B7FF1").unwrap();
+    ///     let m = SimpleModule::new("test_app.pdb", debug_id);
+    ///     let mut f = SimpleFrame::with_instruction(0x1010);
+    ///     let _ = symbolizer.fill_symbol(&m, &mut f).await;
+    ///     assert_eq!(f.function.unwrap(), "vswprintf");
+    ///     assert_eq!(f.source_file.unwrap(),
+    ///         r"c:\program files\microsoft visual studio 8\vc\include\swprintf.inl");
+    ///     assert_eq!(f.source_line.unwrap(), 51);
+    /// }
     /// ```
     ///
     /// [simplemodule]: struct.SimpleModule.html
     /// [simpleframe]: struct.SimpleFrame.html
-    pub fn fill_symbol(
+    pub async fn fill_symbol(
         &self,
-        module: &dyn Module,
-        frame: &mut dyn FrameSymbolizer,
+        module: &(dyn Module + Sync),
+        frame: &mut (dyn FrameSymbolizer + Send),
     ) -> Result<(), FillSymbolError> {
         let k = key(module);
-        self.ensure_module(module, &k);
+        self.ensure_module(module, &k).await;
 
         // Symbols will always contain an entry after ensure_module (though it may be an Err).
-        self.symbols.borrow()[&k]
+        self.symbols.lock().unwrap()[&k]
             .as_ref()
             .map(|sym| {
                 sym.fill_symbol(module, frame);
@@ -662,7 +738,8 @@ impl Symbolizer {
     /// Keys are the file name of the module (code_file's file name).
     pub fn stats(&self) -> HashMap<String, SymbolStats> {
         self.symbols
-            .borrow()
+            .lock()
+            .unwrap()
             .iter()
             .map(|(k, res)| {
                 let mut stats = SymbolStats::default();
@@ -675,10 +752,13 @@ impl Symbolizer {
                     Err(SymbolError::NotFound) => {
                         stats.loaded_symbols = false;
                     }
+                    Err(SymbolError::MissingDebugFileOrId) => {
+                        stats.loaded_symbols = false;
+                    }
                     Err(SymbolError::LoadError(_)) => {
                         stats.loaded_symbols = false;
                     }
-                    Err(SymbolError::ParseError(_)) => {
+                    Err(SymbolError::ParseError(..)) => {
                         stats.loaded_symbols = true;
                         stats.corrupt_symbols = true;
                     }
@@ -691,10 +771,14 @@ impl Symbolizer {
     /// Tries to use CFI to walk the stack frame of the FrameWalker
     /// using the symbols of the given Module. Output will be written
     /// using the FrameWalker's `set_caller_*` APIs.
-    pub fn walk_frame(&self, module: &dyn Module, walker: &mut dyn FrameWalker) -> Option<()> {
+    pub async fn walk_frame(
+        &self,
+        module: &(dyn Module + Sync),
+        walker: &mut (dyn FrameWalker + Send),
+    ) -> Option<()> {
         let k = key(module);
-        self.ensure_module(module, &k);
-        if let Some(Ok(ref sym)) = self.symbols.borrow().get(&k) {
+        self.ensure_module(module, &k).await;
+        if let Some(Ok(ref sym)) = self.symbols.lock().unwrap().get(&k) {
             trace!("unwind: found symbols for address, searching for cfi entries");
             sym.walk_frame(module, walker)
         } else {
@@ -706,10 +790,10 @@ impl Symbolizer {
     /// Ensures there is an entry in the `symbols` map for the given key
     /// (although it may be an Error). Will not change the entry if it already
     /// exists (so if they first time we look is an Error, it always will be).
-    fn ensure_module(&self, module: &dyn Module, k: &ModuleKey) {
-        if !self.symbols.borrow().contains_key(k) {
-            let res = self.supplier.locate_symbols(module);
-            self.symbols.borrow_mut().insert(k.clone(), res);
+    async fn ensure_module(&self, module: &(dyn Module + Sync), k: &ModuleKey) {
+        if !self.symbols.lock().unwrap().contains_key(k) {
+            let res = self.supplier.locate_symbols(module).await;
+            self.symbols.lock().unwrap().insert(k.clone(), res);
         }
     }
 }
@@ -750,32 +834,33 @@ mod test {
     use std::fs::File;
     use std::io::Write;
     use std::path::{Path, PathBuf};
-    use tempdir::TempDir;
+    use std::str::FromStr;
 
-    #[test]
-    fn test_relative_symbol_path() {
-        let m = SimpleModule::new("foo.pdb", "abcd1234");
+    #[tokio::test]
+    async fn test_relative_symbol_path() {
+        let debug_id = DebugId::from_str("abcd1234-abcd-1234-abcd-abcd12345678-a").unwrap();
+        let m = SimpleModule::new("foo.pdb", debug_id);
         assert_eq!(
             &relative_symbol_path(&m, "sym").unwrap(),
-            "foo.pdb/abcd1234/foo.sym"
+            "foo.pdb/ABCD1234ABCD1234ABCDABCD12345678a/foo.sym"
         );
 
-        let m2 = SimpleModule::new("foo.pdb", "abcd1234");
+        let m2 = SimpleModule::new("foo.pdb", debug_id);
         assert_eq!(
             &relative_symbol_path(&m2, "bar").unwrap(),
-            "foo.pdb/abcd1234/foo.bar"
+            "foo.pdb/ABCD1234ABCD1234ABCDABCD12345678a/foo.bar"
         );
 
-        let m3 = SimpleModule::new("foo.xyz", "abcd1234");
+        let m3 = SimpleModule::new("foo.xyz", debug_id);
         assert_eq!(
             &relative_symbol_path(&m3, "sym").unwrap(),
-            "foo.xyz/abcd1234/foo.xyz.sym"
+            "foo.xyz/ABCD1234ABCD1234ABCDABCD12345678a/foo.xyz.sym"
         );
 
-        let m4 = SimpleModule::new("foo.xyz", "abcd1234");
+        let m4 = SimpleModule::new("foo.xyz", debug_id);
         assert_eq!(
             &relative_symbol_path(&m4, "bar").unwrap(),
-            "foo.xyz/abcd1234/foo.xyz.bar"
+            "foo.xyz/ABCD1234ABCD1234ABCDABCD12345678a/foo.xyz.bar"
         );
 
         let bad = SimpleModule::default();
@@ -788,35 +873,36 @@ mod test {
         assert!(relative_symbol_path(&bad2, "sym").is_none());
 
         let bad3 = SimpleModule {
-            debug_id: Some("foo".to_string()),
+            debug_id: Some(debug_id),
             ..SimpleModule::default()
         };
         assert!(relative_symbol_path(&bad3, "sym").is_none());
     }
 
-    #[test]
-    fn test_relative_symbol_path_abs_paths() {
+    #[tokio::test]
+    async fn test_relative_symbol_path_abs_paths() {
+        let debug_id = DebugId::from_str("abcd1234-abcd-1234-abcd-abcd12345678-a").unwrap();
         {
-            let m = SimpleModule::new("/path/to/foo.bin", "abcd1234");
+            let m = SimpleModule::new("/path/to/foo.bin", debug_id);
             assert_eq!(
                 &relative_symbol_path(&m, "sym").unwrap(),
-                "foo.bin/abcd1234/foo.bin.sym"
+                "foo.bin/ABCD1234ABCD1234ABCDABCD12345678a/foo.bin.sym"
             );
         }
 
         {
-            let m = SimpleModule::new("c:/path/to/foo.pdb", "abcd1234");
+            let m = SimpleModule::new("c:/path/to/foo.pdb", debug_id);
             assert_eq!(
                 &relative_symbol_path(&m, "sym").unwrap(),
-                "foo.pdb/abcd1234/foo.sym"
+                "foo.pdb/ABCD1234ABCD1234ABCDABCD12345678a/foo.sym"
             );
         }
 
         {
-            let m = SimpleModule::new("c:\\path\\to\\foo.pdb", "abcd1234");
+            let m = SimpleModule::new("c:\\path\\to\\foo.pdb", debug_id);
             assert_eq!(
                 &relative_symbol_path(&m, "sym").unwrap(),
-                "foo.pdb/abcd1234/foo.sym"
+                "foo.pdb/ABCD1234ABCD1234ABCDABCD12345678a/foo.sym"
             );
         }
     }
@@ -848,67 +934,88 @@ mod test {
         write_symbol_file(path, b"this is not a symbol file\n");
     }
 
-    #[test]
-    fn test_simple_symbol_supplier() {
-        let t = TempDir::new("symtest").unwrap();
+    #[tokio::test]
+    async fn test_simple_symbol_supplier() {
+        let t = tempfile::tempdir().unwrap();
         let paths = mksubdirs(t.path(), &["one", "two"]);
 
         let supplier = SimpleSymbolSupplier::new(paths.clone());
         let bad = SimpleModule::default();
-        assert_eq!(supplier.locate_symbols(&bad), Err(SymbolError::NotFound));
+        assert_eq!(
+            supplier.locate_symbols(&bad).await,
+            Err(SymbolError::NotFound)
+        );
 
         // Try loading symbols for each of two modules in each of the two
         // search paths.
         for &(path, file, id, sym) in [
-            (&paths[0], "foo.pdb", "abcd1234", "foo.pdb/abcd1234/foo.sym"),
-            (&paths[1], "bar.xyz", "ff9900", "bar.xyz/ff9900/bar.xyz.sym"),
+            (
+                &paths[0],
+                "foo.pdb",
+                DebugId::from_str("abcd1234-0000-0000-0000-abcd12345678-a").unwrap(),
+                "foo.pdb/ABCD1234000000000000ABCD12345678a/foo.sym",
+            ),
+            (
+                &paths[1],
+                "bar.xyz",
+                DebugId::from_str("ff990000-0000-0000-0000-abcd12345678-a").unwrap(),
+                "bar.xyz/FF990000000000000000ABCD12345678a/bar.xyz.sym",
+            ),
         ]
         .iter()
         {
             let m = SimpleModule::new(file, id);
             // No symbols present yet.
-            assert_eq!(supplier.locate_symbols(&m), Err(SymbolError::NotFound));
+            assert_eq!(
+                supplier.locate_symbols(&m).await,
+                Err(SymbolError::NotFound)
+            );
             write_good_symbol_file(&path.join(sym));
             // Should load OK now that it exists.
             assert!(
-                matches!(supplier.locate_symbols(&m), Ok(_)),
+                matches!(supplier.locate_symbols(&m).await, Ok(_)),
                 "{}",
                 format!("Located symbols for {}", sym)
             );
         }
 
         // Write a malformed symbol file, verify that it's found but fails to load.
-        let mal = SimpleModule::new("baz.pdb", "ffff0000");
-        let sym = "baz.pdb/ffff0000/baz.sym";
-        assert_eq!(supplier.locate_symbols(&mal), Err(SymbolError::NotFound));
+        let debug_id = DebugId::from_str("ffff0000-0000-0000-0000-abcd12345678-a").unwrap();
+        let mal = SimpleModule::new("baz.pdb", debug_id);
+        let sym = "baz.pdb/FFFF0000000000000000ABCD12345678a/baz.sym";
+        assert_eq!(
+            supplier.locate_symbols(&mal).await,
+            Err(SymbolError::NotFound)
+        );
         write_bad_symbol_file(&paths[0].join(sym));
-        let res = supplier.locate_symbols(&mal);
+        let res = supplier.locate_symbols(&mal).await;
         assert!(
-            matches!(res, Err(SymbolError::ParseError(_))),
+            matches!(res, Err(SymbolError::ParseError(..))),
             "{}",
             format!("Correctly failed to parse {}, result: {:?}", sym, res)
         );
     }
 
-    #[test]
-    fn test_symbolizer() {
-        let t = TempDir::new("symtest").unwrap();
+    #[tokio::test]
+    async fn test_symbolizer() {
+        let t = tempfile::tempdir().unwrap();
         let path = t.path();
 
         // TODO: This could really use a MockSupplier
         let supplier = SimpleSymbolSupplier::new(vec![PathBuf::from(path)]);
         let symbolizer = Symbolizer::new(supplier);
-        let m1 = SimpleModule::new("foo.pdb", "abcd1234");
+        let debug_id = DebugId::from_str("abcd1234-abcd-1234-abcd-abcd12345678-a").unwrap();
+        let m1 = SimpleModule::new("foo.pdb", debug_id);
         write_symbol_file(
-            &path.join("foo.pdb/abcd1234/foo.sym"),
-            b"MODULE Linux x86 abcd1234 foo
+            &path.join("foo.pdb/ABCD1234ABCD1234ABCDABCD12345678a/foo.sym"),
+            b"MODULE Linux x86 ABCD1234ABCD1234ABCDABCD12345678a foo
 FILE 1 foo.c
 FUNC 1000 30 10 some func
 1000 30 100 1
 ",
         );
         let mut f1 = SimpleFrame::with_instruction(0x1010);
-        symbolizer.fill_symbol(&m1, &mut f1).unwrap();
+        symbolizer.fill_symbol(&m1, &mut f1).await.unwrap();
         assert_eq!(f1.function.unwrap(), "some func");
         assert_eq!(f1.function_base.unwrap(), 0x1000);
         assert_eq!(f1.source_file.unwrap(), "foo.c");
@@ -917,36 +1024,39 @@ FUNC 1000 30 10 some func
 
         assert_eq!(
             symbolizer
-                .get_symbol_at_address("foo.pdb", "abcd1234", 0x1010)
+                .get_symbol_at_address("foo.pdb", debug_id, 0x1010)
+                .await
                 .unwrap(),
             "some func"
         );
 
-        let m2 = SimpleModule::new("bar.pdb", "ffff0000");
+        let debug_id = DebugId::from_str("ffff0000-0000-0000-0000-abcd12345678-a").unwrap();
+        let m2 = SimpleModule::new("bar.pdb", debug_id);
         let mut f2 = SimpleFrame::with_instruction(0x1010);
         // No symbols present, should not find anything.
-        assert!(symbolizer.fill_symbol(&m2, &mut f2).is_err());
+        assert!(symbolizer.fill_symbol(&m2, &mut f2).await.is_err());
         assert!(f2.function.is_none());
         assert!(f2.function_base.is_none());
         assert!(f2.source_file.is_none());
         assert!(f2.source_line.is_none());
         // Results should be cached.
         write_symbol_file(
-            &path.join("bar.pdb/ffff0000/bar.sym"),
-            b"MODULE Linux x86 ffff0000 bar
+            &path.join("bar.pdb/ffff0000000000000000ABCD12345678a/bar.sym"),
+            b"MODULE Linux x86 ffff0000000000000000ABCD12345678a bar
 FILE 53 bar.c
 FUNC 1000 30 10 another func
 1000 30 7 53
 ",
         );
-        assert!(symbolizer.fill_symbol(&m2, &mut f2).is_err());
+        assert!(symbolizer.fill_symbol(&m2, &mut f2).await.is_err());
         assert!(f2.function.is_none());
         assert!(f2.function_base.is_none());
         assert!(f2.source_file.is_none());
         assert!(f2.source_line.is_none());
         // This should also use cached results.
         assert!(symbolizer
-            .get_symbol_at_address("bar.pdb", "ffff0000", 0x1010)
+            .get_symbol_at_address("bar.pdb", debug_id, 0x1010)
+            .await
             .is_none());
     }
 }
